import boto3
import collections
import glob
import logging
import inspect
import json
import os
import shutil
import tempfile
import time
import types
import typing
import zipfile

from bert.deploy import shortcuts as bert_deploy_shortcuts

from botocore.errorfactory import ClientError

# Related: https://github.com/Miserlou/Zappa/pull/56
# Related: https://github.com/Miserlou/Zappa/pull/581
ZIP_EXCLUDES: typing.List[str] = [
    '*.exe', '*.DS_Store', '*.Python', '*.git', '.git/*', '*.zip', '*.tar.gz',
    '*.hg', 'pip', 'docutils*', 'setuputils*', '__pycache__/*',
]
COMMON_EXCLUDES: typing.List[str] = ['env', 'lambdas']

logger = logging.getLogger(__name__)

def _calc_lambda_name(lambda_name: str) -> str:
    return lambda_name

def _calc_table_name(lambda_key: str) -> str:
    return lambda_key

def copytree(src: str, dest: str, metadata: bool = True, symlinks: bool = False, ignore: typing.Any = None) -> None:
    if not os.path.exists(dest):
        os.makedirs(dest)
        if metadata:
            shutil.copystat(src, dest)

    dir_list: typing.List[str] = os.listdir(src)
    if ignore:
        excl = ignore(src, dir_list)
        dir_list = [item for item in dir_list if item not in excl]

    for item in dir_list:
        source_path: str = os.path.join(src, item)
        dest_path: str = os.path.join(dest, item)
        if symlinks and os.path.islink(source_path):
            if os.path.lexsits(dest_path):
                os.remove(dest_path)

            os.symlink(os.readlink(source_path), dest_path)
            if metadata:
                try:
                    stats: typing.Any = os.lstat(source_path)
                    mode = stat.S_IMODE(stats.st_mode)
                    os.lchmod(dest_path, mode)
                except:
                    pass # lchmod not availabe

        elif os.path.isdir(source_path):
            copytree(source_path, dest_path, metadata, symlinks, ignore)

        else:
            shutil.copy2(source_path, dest_path) if metadata else shutil.copy(source_path, dest_path)

def find_site_packages_dir(start: str, find: str) -> str:
    """
    Scan a directory[start] for a subdirectory[find]
    """
    if not os.path.isdir(start):
        raise IOError(f'Path[{start}] is not a directory.')

    for root, dirs, files in os.walk(start):
        if find in dirs:
            return os.path.join(start, find)

        for dir_name in dirs:
            result: str = find_site_packages_dir(os.path.join(start, dir_name), find)
            if result:
                return result

def get_current_venv(excludes: typing.List[str] = []):
    if not 'VIRTUAL_ENV' in os.environ.keys():
        return None

    venv = os.environ['VIRTUAL_ENV']
    # egg_links: typing.List[str] = []
    temp_package_path: str = tempfile.mkdtemp(prefix='bert-etl-packages')
    site_package_dir: str = find_site_packages_dir(venv, 'site-packages')
    # egg_links.extend(glob.glob(os.path.join(site_package_dir, '*.egg-link')))

    excludes = ZIP_EXCLUDES + excludes + ['lamdbas']
    logger.info(f'Creating Site-Packages Path[{temp_package_path}]')
    copytree(site_package_dir, temp_package_path, metadata=False, symlinks=False, ignore=shutil.ignore_patterns(*excludes))
    return temp_package_path

def get_pyenv_venv():
    if not os.path.exists('.python-version'):
        return None

def get_conda_venv(excludes: typing.List[str] = []):
    from distutils.sysconfig import get_python_lib
    python_lib: str = get_python_lib()
    if 'conda/envs' in python_lib or 'miniconda/envs' in python_lib:
        temp_package_path: str = tempfile.mkdtemp(prefix='bert-etl-packages')
        excludes = ZIP_EXCLUDES + excludes + ['lamdbas']
        logger.info(f'Creating Site-Packages Path[{temp_package_path}]')
        copytree(python_lib, temp_package_path, metadata=False, symlinks=False, ignore=shutil.ignore_patterns(*excludes))
        return temp_package_path

def build_project_envs(jobs: typing.Dict[str, types.FunctionType], venv_path: str, excludes: typing.List[str] = COMMON_EXCLUDES) -> typing.Dict[str, typing.Any]:
    confs: typing.Dict[str, typing.Any] = {}
    bert_configuration = bert_deploy_shortcuts.load_local_configuration()

    for job_name, job in jobs.items():
        env_vars: typing.Dict[str, str] = bert_deploy_shortcuts.merge_env_vars(
            bert_configuration.get('every_lambda', {'environment': {}})['environment'],
            bert_configuration.get(job_name, {'environment': {}})['environment'])
        env_vars['QUEUE_TYPE'] = env_vars.get('QUEUE_TYPE', 'dynamodb')
        runtime: str = bert_configuration.get('runtime', 'python3.6')
        try:
            memory_size: str = int(bert_configuration.get('memory_size', '512'))
        except ValueError:
            raise NotImplementedError('MemorySize must be an integer')

        project_path: str = tempfile.mkdtemp(prefix=f'bert-etl-project-{job_name}')
        logger.info(f'Creating Project Path[{project_path}]')
        job_templates: str = ''.join([f"""def {jn}():
    pass
""" for jn in jobs.keys() if jn != job_name])

        job_follow: str = inspect.getsource(job).split('\n')[0]
        job_source: str = '\n'.join(inspect.getsource(job).split('\n')[2:])
        job_template: str = """
%s
import typing
from bert import utils, constants, binding, shortcuts
%s
def %s(event, context=None):

    records: typing.List[typing.Dict[str, typing.Any]] = event.get('Records', [])
    if len(records) > 0:
        constants.QueueType = constants.QueueTypes.StreamingQueue
        work_queue, done_queue, ologger = utils.comm_binders(%s)
        for record in records:
            if record['eventName'].lower() == 'INSERT'.lower():
                work_queue.local_put(record['dynamodb']['NewImage'])

    else:
        work_queue, done_queue, ologger = utils.comm_binders(%s)

%s
    return {}
""" % (job_templates, job_follow, job_name, job_name, job_name, job_source)
        job_path: str = os.path.join(project_path, f'{job_name}.py')

        with open(job_path, 'w') as stream:
            stream.write(job_template)

        logger.info(f'Creating Job[{job_name}] Project')
        copytree(os.getcwd(), project_path, metadata=False, symlinks=False, ignore=shutil.ignore_patterns(*excludes))
        logger.info(f'Merging Job[{job_name}] Site Packages')
        copytree(venv_path, project_path, metadata=False, symlinks=False, ignore=shutil.ignore_patterns(*excludes))
        confs[_calc_lambda_name(job_name)] = {
                'project-path': project_path,
                'table-name': f'{job_name}-stream',
                # 'runtime': 'python3.7',
                'runtime': runtime,
                'memory-size': memory_size, # must be a multiple of 64, increasing memory size also increases cpu allocation
                'environment': env_vars,
                'handler-name': f'{job_name}.{job_name}',
                'spaces': {
                    'work-key': job.work_key,
                    'done-key': job.done_key,
                    'pipeline-type': job.pipeline_type,
                    'workers': job.workers,
                    'scheme': job.schema,
                }
            }

    return confs

def build_package(job_name: str, job_conf: typing.Dict[str, typing.Any], excludes: typing.List[str] = COMMON_EXCLUDES) -> None:
    try:
        compression_method: int = zipfile.ZIP_DEFLATED
    except ImportError: #pragma: no cover
        compression_method: int = zipfile.ZIP_STORED

    archive_name: str = f'{job_name}.zip'
    archive_dir: str = os.path.join(os.getcwd(), 'lambdas')
    if not os.path.exists(archive_dir):
        os.makedirs(archive_dir)

    archive_path: str = os.path.join(archive_dir, archive_name)
    logger.info(f'Building Lambda[{job_name}] Archive[{archive_path}]')

    job_conf['archive-path'] = archive_path
    with zipfile.ZipFile(archive_path, 'w', compression_method) as archive:
        for root, dirs, files in os.walk(job_conf['project-path']):
            for filename in files:
                if filename in excludes:
                    continue

                if filename.endswith('.pyc'):
                    continue

                abs_filename: str = os.path.join(root, filename)
                if filename.endswith('.py'):
                    os.chmod(abs_filename, 0o755)

                zip_info: zipfile.ZipInfo = zipfile.ZipInfo(os.path.join(root.replace(job_conf['project-path'], '').lstrip(os.sep), filename))
                zip_info.create_system = 3
                zip_info.external_attr = 0o755 << int(16)
                with open(abs_filename, 'rb') as file_stream:
                    archive.writestr(zip_info, file_stream.read(), compression_method)

            for dirname in dirs:
                if dirname in excludes:
                    continue


    job_conf['archive-path'] = archive_path

def build_lambda_archives(jobs: typing.Dict[str, types.FunctionType]) -> str:
    venv_path: str = get_current_venv()
    if venv_path is None:
        venv_path = get_pyenv_venv()
        if venv_path is None:
            venv_path = get_conda_venv()
            if venv_path is None:
                raise NotImplementedError

    archive_paths: typing.List[str] = []
    lambdas: typing.Dict[str, typing.Any] = collections.OrderedDict()
    for job_name, job_conf in build_project_envs(jobs, venv_path).items():
        build_package(job_name, job_conf)
        lambdas[job_name] = job_conf
        # shutil.rmtree(venv_path)
        shutil.rmtree(job_conf['project-path'])

    return lambdas

def destroy_dynamodb_tables(jobs: typing.Dict[str, typing.Any]) -> None:
    table: typing.Any = None
    client = boto3.client('dynamodb')
    for job_name, job in jobs.items():
        work_table_name: str = _calc_table_name(job.work_key)
        done_table_name: str = _calc_table_name(job.done_key)

        try:
            client.delete_table(TableName=work_table_name)
        except ClientError as err:
            pass

        else:
            logger.info(f'Destorying Table[{work_table_name}]')

        try:
            client.delete_table(TableName=done_table_name)
        except ClientError as err:
            pass

        else:
            logger.info(f'Destorying Table[{done_table_name}]')


def build_dynamodb_tables(lambdas: typing.Dict[str, typing.Any]) -> None:
    table: typing.Any = None
    client = boto3.client('dynamodb')
    for lambda_name, conf in lambdas.items():
        work_table_name: str = _calc_table_name(conf['spaces']['work-key'])
        done_table_name: str = _calc_table_name(conf["spaces"]["done-key"])

        try:
            conf['work-table'] = client.describe_table(TableName=work_table_name)
        except ClientError as err:
            logger.info(f'Creating Dynamodb Table[{work_table_name}]')
            client.create_table(
                    TableName=work_table_name,
                    KeySchema=[
                        {
                            'AttributeName': 'identity',
                            'KeyType': 'HASH'
                        }
                    ],
                    AttributeDefinitions=[
                        {
                            'AttributeName': 'identity',
                            'AttributeType': 'S'
                        }
                    ],
                    StreamSpecification={
                        'StreamEnabled': True,
                        'StreamViewType': 'NEW_IMAGE'
                    },
                    BillingMode='PAY_PER_REQUEST')
            conf['work-table'] = client.describe_table(TableName=work_table_name)

        try:
            conf['done-table'] = client.describe_table(TableName=done_table_name)
        except ClientError as err:
            logger.info(f'Creating Dynamodb Table[{done_table_name}]')
            client.create_table(
                    TableName=done_table_name,
                    KeySchema=[
                        {
                            'AttributeName': 'identity',
                            'KeyType': 'HASH'
                        }
                    ],
                    AttributeDefinitions=[
                        {
                            'AttributeName': 'identity',
                            'AttributeType': 'S'
                        }
                    ],
                    StreamSpecification={
                        'StreamEnabled': True,
                        'StreamViewType': 'NEW_IMAGE'
                    },
                    BillingMode='PAY_PER_REQUEST')

            conf['done-table'] = client.describe_table(TableName=done_table_name)


def create_lambda_roles(lambdas: typing.Dict[str, typing.Any]) -> None:
    trust_policy_document = {
        "Version": "2012-10-17",
        "Statement": [
            {
                "Sid": "",
                "Effect": "Allow",
                "Principal": {
                    "Service": [
                        "apigateway.amazonaws.com",
                        "lambda.amazonaws.com",
                        "events.amazonaws.com",
                        "dynamodb.amazonaws.com",
                    ]
                },
                "Action": "sts:AssumeRole",
            }
        ]
    }

    policy_document: typing.Dict[str, typing.Any] = {
        "Version": "2012-10-17",
        "Statement": [
            # {
            #     "Effect": "Allow",
            #     "Action": [
            #         "xray:PutTraceSegments",
            #         "xray:PutTelemetryRecords",
            #     ],
            #     "Resource": ["*"],
            # },
            # {
            #     "Effect": "Allow",
            #     "Action": [
            #         "lambda:InvokeFunction"
            #     ],
            #     "Resource": ["*"],
            # },
            {
                "Effect": "Allow",
                "Action": [
                    "dynamodb:*",
                ],
                "Resource": "arn:aws:dynamodb:*:*:table/*",
            },
            {
                "Effect": "Allow",
                "Action": "s3:*",
                "Resource": "*"
            },
            {
                "Effect": "Allow",
                "Action": [
                    "logs:CreateLogStream",
                    "logs:PutLogEvents",
                    "logs:CreateLogGroup",
                ],
                "Resource": "arn:aws:logs:*:*:*",
            }
        ]
    }
    trust_policy_name: str = 'bert-etl-lambda-execution-policy-trust'
    policy_name: str = 'bert-etl-lambda-execution-policy'
    iam_role = {
        'Path': '/',
        'RoleName': trust_policy_name,
        'AssumeRolePolicyDocument': json.dumps(trust_policy_document),
        'Description': 'Bert-ETL Lambda Execution Role',
    }
    iam_policy = {
        'Path': '/',
        'PolicyName': policy_name,
        'PolicyDocument': json.dumps(policy_document),
        'Description': 'Bert-ETL Lambda Execution Policy'
    }
    iam_client = boto3.client('iam')
    role = bert_deploy_shortcuts.map_iam_role(trust_policy_name)
    if role is None:
        iam_client.create_role(**iam_role)
        role = bert_deploy_shortcuts.map_iam_role(trust_policy_name)

    policy = bert_deploy_shortcuts.map_iam_policy(policy_name)
    if policy is None:
        iam_client.create_policy(**iam_policy)
        policy = bert_deploy_shortcuts.map_iam_policy(policy_name)
        iam_client.attach_role_policy(RoleName=trust_policy_name, PolicyArn=policy['Arn'])

    for job_name, conf in lambdas.items():
        conf['iam-role'] = role
        conf['iam-policy'] = policy

def destory_lambda_to_table_bindings(lambdas: typing.Dict[str, typing.Any]) -> None:
    client = boto3.client('lambda')
    for lambda_name, conf in lambdas.items():
        for event_mapping in client.list_event_source_mappings(
                EventSourceArn=conf['work-table']['Table']['LatestStreamArn'],
                FunctionName=lambda_name)['EventSourceMappings']:
            client.delete_event_source_mapping(UUID=event_mapping['UUID'])

def destroy_lambdas(jobs: typing.Dict[str, typing.Any]) -> None:
    client = boto3.client('lambda')
    for job_name, job in jobs.items():
        lambda_name: str = _calc_lambda_name(job_name)
        try:
            client.delete_function(FunctionName=lambda_name)
        except ClientError as err:
            pass

        else:
            logger.info(f'Deleting function[{lambda_name}]')


def upload_lambdas(lambdas: typing.Dict[str, typing.Any]) -> None:
    client = boto3.client('lambda')
    for lambda_name, conf in lambdas.items():
        try:
            client.get_function(FunctionName=lambda_name)['Configuration']
        except ClientError as err:
            logger.info(f'Creating AWSLambda for Job[{lambda_name}]')
            lambda_description = client.create_function(
                FunctionName=lambda_name,
                Runtime=conf['runtime'],
                MemorySize=conf['memory-size'],
                Role=conf['iam-role']['Arn'],
                Handler=conf['handler-name'],
                Code={
                    'ZipFile': open(conf['archive-path'], 'rb').read(),
                },
                Timeout=900,
                Environment={'Variables': conf['environment']},
            )
            conf['aws-lambda'] = client.get_function(FunctionName=lambda_name)['Configuration']

        else:
            logger.info(f'Replacing AWSLambda for Job[{lambda_name}]')
            client.delete_function(FunctionName=lambda_name)
            client.create_function(
                FunctionName=lambda_name,
                Runtime=conf['runtime'],
                MemorySize=conf['memory-size'],
                Role=conf['iam-role']['Arn'],
                Handler=conf['handler-name'],
                Code={
                    'ZipFile': open(conf['archive-path'], 'rb').read(),
                },
                Timeout=900,
                Environment={'Variables': conf['environment']},
            )
            conf['aws-lambda'] = client.get_function(FunctionName=lambda_name)['Configuration']

def bind_lambdas_to_tables(lambdas: typing.Dict[str, typing.Any]) -> None:
    client = boto3.client('lambda')
    for lambda_name, conf in lambdas.items():
        logger.info(f'Mapping Lambda[{lambda_name}] to Work-Table')
        client.create_event_source_mapping(
                EventSourceArn=conf['work-table']['Table']['LatestStreamArn'],
                FunctionName=lambda_name,
                Enabled=True,
                StartingPosition='LATEST')




#!/usr/env/bin python

import boto3
import collections
import copy
import logging
import hashlib
import importlib
import json
import multiprocessing
import os
import redis
import shutil
import subprocess
import time
import typing
import types

import requests

from datetime import datetime

from bert import constants, datasource

from urllib.parse import urlparse, ParseResult

logger = logging.getLogger(__name__)
PWN = typing.TypeVar('PWN')

class QueuePacker(json.JSONEncoder):
  def default(self, obj):
    if isinstance(obj, datetime):
      return obj.strftime(constants.DATETIME_FORMAT)

    return super(QueuePacker, self).default(obj)

class Across:
  '''
  Sometimes checking to see if something exists arcoss processes is required to complete a job. Knowing where it exists is faster than searching for where it might exist in space.
  '''

  def __init__(self) -> None:
    self._conn: datasource.RedisConnection = datasource.RedisConnection.ParseURL(constants.REDIS_URL)

  def __enter__(self, redis_key: str) -> PWN:
    self._key = redis_key
    return self

  def __exit__(self, type, value, tb) -> PWN:
    self._key = None
    return self

  def exists(self) -> bool:
    value: typing.Any = self._conn.client.get(self._key)
    if value:
      return True

    return False

def _find_aws_encoding(datum: typing.Any) -> typing.Dict[str, typing.Any]:
    if isinstance(datum, dict):
        return 'M'

    elif isinstance(datum, list):
        return 'L'

    elif isinstance(datum, bytes):
        return 'B'

    elif isinstance(datum, str):
        return 'S'

    else:
        import ipdb; ipdb.set_trace()
        raise NotImplementedError

def _encode_aws_object(datum: typing.Any) -> typing.Dict[str, typing.Any]:
    if isinstance(datum, dict):
        for key, value in datum.items():
            datum[key] = {_find_aws_encoding(value): _encode_aws_object(value)}

        return datum

    elif isinstance(datum, (list, tuple, types.GeneratorType)):
        for idx, value in enumerate(datum):
            datum[idx] = {_find_aws_encoding(value): _encode_aws_object(value)}

        return datum

    else:
        return datum

def _decode_aws_object(datum: typing.Dict[str, typing.Any]) -> typing.Any:
    for encoding_type, encoded in datum.items():
        if encoding_type == 'M':
            for key, value, in encoded.items():
                encoded[key] = _decode_aws_object(value)

            return encoded

        elif encoding_type == 'L':
            for idx, value in enumerate(encoded):
                encoded[idx] = _decode_aws_object(value)

            return encoded

        elif encoding_type == 'B':
            return encoded

        elif encoding_type == 'S':
            return encoded

        else:
            import ipdb; ipdb.set_trace()
            raise NotImplementedError

class DynamodbQueue:
    DEFAULT_DELAY: int = 1728000
    @staticmethod
    def Pack(datum: typing.Dict[str, typing.Any]) -> str:
        """
        Dict -> AWS Dict -> json.dumps
        """
        _encode_aws_object(datum)
        return datum

    @staticmethod
    def UnPack(datum: str) -> typing.Dict[str, typing.Any]:
        """
        AWS Dynamodb Dict -> Dict
        """
        return _decode_aws_object(datum)

    _key: str = None
    def __init__(self: PWN, key: str) -> None:
        self._key = key

    def __iter__(self) -> PWN:
        return self

    def __next__(self) -> typing.Any:
        value = self.get()
        if value is None or value == 'STOP':
            raise StopIteration

        return value

    def put(self: PWN, value: typing.Dict[str, typing.Any]) -> None:
        combined: str = ''.join(sorted(json.dumps(value)))
        identity: str = hashlib.sha256(combined.encode('utf-8')).hexdigest()
        value: str = self.__class__.Pack({'identity': identity, 'datum': copy.deepcopy(value)})
        client: typing.Any = boto3.client('dynamodb')
        client.put_item(TableName=self._key, Item=value)

    def get(self: PWN) -> typing.Dict[str, typing.Any]:
        client: typing.Any = boto3.client('dynamodb')
        try:
            value: typing.Any = client.scan(TableName=self._key, Select='ALL_ATTRIBUTES', Limit=1)['Items'][0]
        except IndexError:
            return None

        else:
            unpacked: typing.Dict[str, typing.Any] = self.__class__.UnPack(copy.deepcopy(value)['datum'])
            client.delete_item(TableName=self._key, Key={'identity': value['identity']})
            return unpacked

class StreamingQueue(DynamodbQueue):
    """
    When deploying functions to AWS Lambda, auto-invocation is available as an option to run the functions. With StreamingQueue, we want to push local objects into
        the available API already utilized. We also want to keep the available `put` function so that the `done_queue` api will still push contents into the next `work_queue`.
        We'll also argment the local `get` function api and only pull from records local to the stream and not pull from dynamodb.
    """
    _key: str = None
    # Share the memory across invocations, within the same process/thread. This allows for
    #   comm_binders to be called multipule-times and still pull from the same queue
    _queue: typing.List[typing.Dict[str, typing.Any]] = []
    def __init__(self: PWN, key: str) -> None:
        self._key = key

    def local_put(self: PWN, record: typing.Dict[str, typing.Any]) -> None:
        print('Putting Record')
        print(json.dumps(record))
        self._queue.append(copy.deepcopy(record))
        print('queue len', id(self), len(self._queue))

    def get(self: PWN) -> typing.Dict[str, typing.Any]:
        print('Getting Record')
        print('queue len', id(self), len(self._queue))
        try:
            value: typing.Any = self._queue.pop(0)
        except IndexError:
            return None

        else:
            print('value')
            print(value)
            unpacked: typing.Dict[str, typing.Any] = self.__class__.UnPack(value['datum'])
            print(unpacked)
            client = boto3.client('dynamodb')
            client.delete_item(TableName=self._key, Key={'identity': value['identity']})
            return unpacked

class Queue:
  DEFAULT_DELAY: int = 1728000
  @staticmethod
  def Pack(datum: typing.Dict[str, typing.Any]) -> str:
    return json.dumps(datum, cls=QueuePacker)

  @staticmethod
  def UnPack(datum: str) -> typing.Dict[str, typing.Any]:
    return json.loads(datum)

  def __init__(self, redis_key: str):
    self._key = redis_key
    self._redis_client = datasource.RedisConnection.ParseURL(constants.REDIS_URL).client

  def __iter__(self):
    return self

  def __next__(self):
    value = self.get()
    if value is None or value == 'STOP':
      raise StopIteration

    return value

  def peg(self, key: str, delay: str=DEFAULT_DELAY) -> bool:
    value: str = self._redis_client.get(key)
    if value is None:
      self._redis_client.set(key, 1)
      self._redis_client.expire(key, delay)
      return True

    return False

  def flushdb(self) -> None:
    self._redis_client.flushdb()

  def put_block(self, value: str, delay: int=1728000) -> None:
    value_sorted = ''.join(sorted(value))
    value_sorted = ''.join(['put-block', value_sorted])
    value_hash = hashlib.md5(value_sorted.encode('utf-8')).hexdigest()
    if self._redis_client.get(value_hash) is None:
      self._redis_client.set(value_hash, 1)
      self._redis_client.expire(value_hash, delay)
      self.put(value)

  def size(self) -> int:
    return int(self._redis_client.llen(self._key))

  def get(self) -> typing.Dict[str, typing.Any]:
    try:
      value: str = self._redis_client.lpop(self._key).decode('utf-8')
    except AttributeError:
      return 'STOP'

    else:
      return self.__class__.UnPack(value)

  def put(self, value: typing.Dict[str, typing.Any]) -> None:
    value: str = self.__class__.Pack(value)

    self._redis_client.rpush(self._key, value.encode('utf-8'))

  def bulk(self, stop_iter_key: str, bulk_count: int=50000) -> typing.List[str]:

    index: int = 0
    datums: typing.List[str] = []
    while index < bulk_count:
      datum: str = self._redis_client.lpop(self._key)
      if datum is None:
        break

      if datum == stop_iter_key:
        break

      datums.append(datum.decode('utf-8'))

    logging.info(len(datums))
    return datums

def scan_jobs(options) -> typing.Dict[str, typing.Any]:
    jobs: typing.Dict[str, typing.Any] = {}
    module = importlib.import_module(f'{options.module_name}.jobs')
    for member_name in dir(module):
        if member_name.startswith('_'):
            continue

        member = getattr(module, member_name)
        if type(member) != types.FunctionType:
            continue

        if not hasattr(member, 'done_key') or not hasattr(member, 'work_key'):
            continue

        jobs[member_name] = member

    # Order the jobs correctly
    ordered = collections.OrderedDict()
    while len(ordered.keys()) != len(jobs.keys()):
        if len(ordered.keys()) == 0:
            for job_name, job in jobs.items():
                if job.parent_func == 'noop':
                    ordered[job_name] = job
                    break

            else:
                raise NotImplementedError(f'NoopSpace not found')

        else:
            latest: types.FunctionType = [item for item in ordered.values()][-1]
            for job_name, job in jobs.items():
                if job.parent_func == latest:
                    ordered[job_name] = job
                    break
    return ordered

def find_in_elem(elem, *args) -> str:
  base = elem
  for key in args:
    try:
      base = base.find(fix_tag(key))
    except Exception as err:
      return 0

  else:
    if base == None:
      return 0

    elif base.text == None:
      return 0

    elif isinstance(base.text, bytes):
      return base.text.decode('utf-8')

    elif isinstance(base.text, str):
      return base.text

    else:
      import ipdb; ipdb.set_trace()
      return 0

def parse_timestamp(input_datetime: str) -> datetime:
  if input_datetime == 0:
    return None

  return datetime.strptime(input_datetime, '%Y-%m-%dT%H:%M:%SZ')

def fix_schema_tag(tag) -> str:
  return '{http://www.w3.org/2001/XMLSchema}%s' % tag

def fix_tag(tag) -> str:
  return '{http://www.mediawiki.org/xml/export-0.10/}%s' % tag

def zero_file(filepath, file_size):
  with open(filepath, 'wb') as stream:
    for chunk in iter(lambda: b'\0' * 4096, b''):
      stream.write(chunk)
      if stream.tell() > file_size:
        break

def run_command(cmd: str, allow_error: typing.List[int] = [0]) -> str:
  ologger = logging.getLogger('.'.join([__name__, multiprocessing.current_process().name]))
  cmd: typing.List[str] = cmd.split(' ')
  proc = subprocess.Popen(' '.join(cmd), stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
  while proc.poll() is None:
    time.sleep(.1)

  if proc.poll() > 0:
    if not proc.poll() in allow_error:
      raise NotImplementedError(f'{proc.poll()}, {proc.stderr.read()}')

    return proc.stderr.read().decode(constants.ENCODING)

  return proc.stdout.read().decode('utf-8')

def comm_binders(func: types.FunctionType) -> typing.Tuple[Queue, Queue, 'ologger']:
    logger.info(f'Bert Queue Type[{constants.QueueType}]')
    ologger = logging.getLogger('.'.join([func.__name__, multiprocessing.current_process().name]))
    if constants.QueueType is constants.QueueTypes.Dynamodb:
        return DynamodbQueue(func.work_key), DynamodbQueue(func.done_key), ologger

    elif constants.QueueType is constants.QueueTypes.StreamingQueue:
        return StreamingQueue(func.work_key), StreamingQueue(func.done_key), ologger

    # elif constants.QueueType is constants.QueueTypes.DynamodbStreamBottle:
    #     return DynamodbStreamQueue(func.work_key), DynamodbStreamBottleQueue(func.done_key), ologger

    elif constants.QueueType is constants.QueueTypes.Redis:
        return Queue(func.work_key), Queue(func.done_key), ologger

    else:
        raise NotImplementedError(constants.QueueType)

def new_module(options: typing.Any) -> None:
  import bert
  operating_dir: str = os.path.join(os.getcwd(), options.new_module)
  if not os.path.exists(operating_dir):
    os.makedirs(operating_dir)

  bert_jobs_filepath: str = os.path.join(os.path.dirname(bert.__file__), 'jobs.py')
  jobs_file: str = os.path.join(operating_dir, 'jobs.py')
  if not os.path.exists(jobs_file):
    shutil.copyfile(bert_jobs_filepath, jobs_file)

  logger.info(f'''
  Run the new module with the following command:
  bert-runner.py -j sync_sounds -m {options.new_module}
  ''')

def show_how(options: typing.Any) -> None:
  logger.info('''
  Run the new module with the following command:
  bert-runner.py -j <job_name> -m <module_name>
  ''')




import logging
import os
import sys

logger = logging.getLogger(__name__)
from bert import constants, remote_webservice
logger.info(f'Starting service[{constants.SERVICE_NAME}] Daemon. Debug[{constants.DEBUG}]')
logger.info(f'Loading Service Module[{constants.SERVICE_MODULE}]')
if constants.SERVICE_MODULE is None:
  raise NotImplementedError(f'Missing ENVVar[SERVICE_MODULE]')

remote_webservice.load_service_module(constants.SERVICE_MODULE)
MIDDLEWARE = remote_webservice.setup_service()



import boto3
import logging
import os
import typing
import yaml

logger = logging.getLogger(__name__)

def merge_env_vars(defaults: typing.Dict[str, str], env_vars: typing.Dict[str, str]) -> typing.Dict[str, str]:
    merged: typing.Dict[str, str] = {key: value for key, value in defaults.items()}
    for key, value in env_vars.items():
        merged[key] = value

    return merged

def load_local_configuration() -> typing.Dict[str, typing.Any]:
    conf_path: str = os.path.join(os.getcwd(), 'bert-etl.yaml')
    if not os.path.exists(conf_path):
        logger.info("Configuration Path[{conf_path}] doesn't exist")
        return None

    with open(conf_path, 'r', encoding='utf-8') as stream:
        return yaml.load(stream.read(), Loader=yaml.FullLoader)

def aws_account_id() -> str:
    sts_client = boto3.client('sts')
    return sts.get_caller_identity()['Account']

def map_iam_role(role_name: str) -> typing.Dict[str, typing.Any]:
    iam_client = boto3.client('iam')
    for page in iam_client.get_paginator('list_roles').paginate(PathPrefix='/'):
        for role in page['Roles']:
            if role['RoleName'] == role_name:
                return role

def map_iam_policy(policy_name: str) -> typing.Dict[str, typing.Any]:
    iam_client = boto3.client('iam')
    for page in iam_client.get_paginator('list_policies').paginate(PathPrefix='/'):
        for policy in page['Policies']:
            if policy['PolicyName'] == policy_name:
                return policy

